#!/usr/bin/env python

import nsysstats

class NVTXPushPopSummary(nsysstats.Report):

    EVENT_TYPE_NVTX_DOMAIN_CREATE = 75
    EVENT_TYPE_NVTX_PUSHPOP_RANGE = 59
    EVENT_TYPE_NVTXT_PUSHPOP_RANGE = 70

    usage = f"""{{SCRIPT}} -- NVTX Push/Pop Range Trace

    No arguments.

    Output: All time values given in nanoseconds
        Start : Event start timestamp
        End : Event end timestamp
        Duration : Duration of this event
        DurChild : Duration of all child events
        DurNonChild : Duration of this event minus child events
        Name : Name of NVTX event
        PID : Process ID
        TID : Thread ID
        Lvl : Stack level, starts at 0
        NumChild : Number of children events
        EventId : Arbitrary ID for event
        ParentId : The event id of the enclosing event
        EventStack : The event ids that make up the push/pop stack
        NameTree : Event name prefixed with level indicator

    This report provides a trace of NV Tools Extensions Push/Pop Ranges,
    their execution time, stack state, and replationship to other push/pop
    events.
"""

# These are the "setup" statements executed before the main query.
# These are not allowed to generate output, other than errors.

    statements = [

# First, create a 1D R-Tree index that will hold NVTX timestamps.  This will
# be used to figure out which events are the children of other events.
# The R-Tree uses 32-bit floating point for its dimentional values, so
# we need two copies of the timestamps-- the indexed floating point values
# to get us close, and the exact values to do a final, detailed check.

f"""
    CREATE VIRTUAL TABLE temp.NVTX_EVENTS_RIDX
    USING rtree
    (
        eventId,
        startTS,
        endTS,
        +startNS   INTEGER,
        +endNS     INTEGER,
        +tid       INTEGER
    );
""",

# Insert NVTX push/pop event data into the R-Tree index
# Not all NVTX events have a valid "end" timestamp, so
# we have to play some games.

f"""
    INSERT INTO temp.NVTX_EVENTS_RIDX
        WITH
            maxts AS(
                SELECT max(max(start), max(end)) AS m
                FROM   NVTX_EVENTS
            )
        SELECT
            e.rowid AS eventId,
            e.start AS startTS,
            ifnull(e.end, (SELECT m FROM maxts)) AS endTS,
            e.start AS startNS,
            ifnull(e.end, (SELECT m FROM maxts)) AS endNS,
            e.globalTid AS tid
        FROM
            NVTX_EVENTS AS e
        WHERE
            e.eventType == {EVENT_TYPE_NVTX_PUSHPOP_RANGE}
            OR
            e.eventType == {EVENT_TYPE_NVTXT_PUSHPOP_RANGE}
    ;
""",

# Create a temp table to hold the parent relationships.
# We're going to compute and hold some meta-data (such as
# number of children, durations, etc.) as well.

f"""
    CREATE TEMP TABLE NVTX_PARENT (
        eventId         INTEGER PRIMARY KEY   NOT NULL,
        parentId        INTEGER,
        duration        INTEGER,
        childDuration   INTEGER,
        childNumb       INTEGER,
        fullname        TEXT
    );
""",

# Insert NVTX push/pop event data into the parent table.
# We do an initial insert of all data, and then run an
# update on the parent data, rather than just inserting
# it all at once, so that the table includes root nodes
# that don't have parents.  It is easier to deal with it
# here than trying to do an OUTER JOIN with the update.

f"""
    INSERT INTO temp.NVTX_PARENT
        WITH
            domains AS (
                SELECT
                    domainId AS id,
                    text AS name
                FROM
                    NVTX_EVENTS
                WHERE
                    eventType == {EVENT_TYPE_NVTX_DOMAIN_CREATE}
            ),
            maxts AS(
                SELECT max(max(start), max(end)) AS m
                FROM   NVTX_EVENTS
            )
        SELECT
            e.rowid AS eventId,
            NULL AS parentId,
            ifnull(e.end, (SELECT m FROM maxts)) - e.start AS duration,
            0 AS childDuration,
            0 AS childNumb,
            CASE
                WHEN d.name NOT NULL AND sid.value IS NOT NULL
                    THEN d.name || ':' || sid.value
                WHEN d.name NOT NULL AND sid.value IS NULL
                    THEN d.name || ':' || e.text
                WHEN d.name IS NULL AND sid.value NOT NULL
                    THEN sid.value
                ELSE e.text
            END AS fullname
        FROM
            NVTX_EVENTS AS e
        LEFT JOIN
            domains AS d
            ON (e.domainId == d.id)
        LEFT JOIN
            StringIds AS sid
            ON (e.textId == sid.id)
        WHERE
            e.eventType == {EVENT_TYPE_NVTX_PUSHPOP_RANGE}
            OR
            e.eventType == {EVENT_TYPE_NVTXT_PUSHPOP_RANGE}
    ;
""",

# Use the R-Tree to figure out which events are children of other events.
# This is done by figuring out which event timestamps are "inside" other
# events and extracting the "tightest" parent.  This tightness is used
# to filter parents from grandparents.  This query depends on a documented,
# but non-standard, behavior of SQLite where the min() aggregate call will
# return the whole row that triggers the minimum, including the corresponding
# event IDs.

f"""
    WITH
        par AS (
            SELECT
                cr.eventId as cid,
                pr.eventId as pid,
                min((cr.startNS - pr.startNS) + (pr.endNS - cr.EndNS)) as tightness
            FROM
                temp.NVTX_EVENTS_RIDX AS cr
            JOIN
                temp.NVTX_EVENTS_RIDX AS pr
            ON
                pr.eventId != cr.eventId
                AND pr.startTS <= cr.startTS
                AND pr.endTS >= cr.endTS
                AND pr.startNS <= cr.startNS
                AND pr.endNS >= cr.endNS
                AND pr.tid == cr.tid
                GROUP BY cid
        )
    UPDATE temp.NVTX_PARENT
        SET parentId = (SELECT pid FROM par WHERE eventId == par.cid)
    ;
""",

# Update the child duration and count

f"""
    WITH
        totals AS (
            SELECT
                parentId AS parentId,
                total(duration) AS childDuration,
                count(*) AS childNumb
            FROM
                NVTX_PARENT
            GROUP BY 1
        )
    UPDATE temp.NVTX_PARENT
        SET (childDuration, childNumb) = (
            SELECT
                childDuration AS childDuration,
                childNumb AS childNumb
            FROM totals
            WHERE totals.parentId == eventId
        )
    ;
""",

# Now that we have our parent data, create an index over the parent ID value.
# When dealing with a fixed data set it is slightly more effecient to create
# the index after all the rows have been inserted.

f"""
    CREATE INDEX IF NOT EXISTS temp.NVTX_PARENT__PARENTID
        ON NVTX_PARENT (parentId)
    ;
""",

] # end of statements

# The actual query uses the table of parents in a recursve CTE to build
# a tree-based query that is aware of what stack level we're on.

    query = f"""
WITH RECURSIVE
    tree AS (
        SELECT
            p.eventId AS eventId,
            ':' || CAST(p.eventId AS TEXT) AS eventIdHier,
            p.parentId AS parentId,
            0 AS level,
            '' AS tab
        FROM
            temp.NVTX_PARENT AS p
        WHERE p.parentId IS NULL

        UNION ALL
        SELECT
            p.eventId AS eventId,
            tree.eventIdHier || ':' || CAST(p.eventId AS TEXT) AS eventIdHier,
            p.parentId AS parentId,
            tree.level + 1 AS level,
            tree.tab || '--' AS tab
        FROM
            tree
        JOIN
            temp.NVTX_PARENT AS p
            ON (p.parentId == tree.eventId)

        ORDER BY level DESC
    )
SELECT
    ne.start AS "Start",
    ne.start + p.duration AS "End",
    p.duration AS "Duration",
    ifnull(p.childDuration, 0) AS "DurChild",
    p.duration - ifnull(p.childDuration, 0) AS "DurNonChild",
    p.fullname AS "Name",
    (ne.globalTid >> 24) & 0x00FFFFFF AS "PID",
    ne.globalTid & 0x00FFFFFF AS "TID",
    t.level AS "Lvl",
    ifnull(p.childNumb, 0) AS "NumChild",
    ne.rowid AS "EventId",
    t.parentId AS "ParentId",
    t.eventIdHier AS "EventStack",
    t.tab || p.fullname AS "NameTree"
FROM
    NVTX_EVENTS AS ne
JOIN
    temp.NVTX_PARENT AS p
    ON (p.eventId = ne.rowid)
JOIN
    tree AS t
    ON (t.eventId == ne.rowid)
;
"""

    table_checks = {
        'NVTX_EVENTS':
            "{DBFILE} does not contain NV Tools Extension (NVTX) data"
    }

if __name__ == "__main__":
    NVTXPushPopSummary.Main()
